use std::ptr::NonNull;
use std::mem;
use std::alloc;

pub struct RawVec<T> {
    pub ptr: NonNull<T>,
    pub cap: usize,
}

impl<T> RawVec<T> {
    pub fn new() -> Self {
        let cap = if mem::size_of::<T>() == 0 { !0 } else { 0 };
        Self { ptr: NonNull::dangling(), cap}
    }

    pub fn grow(&mut self) {
        let elem_size = mem::size_of::<T>();

        assert!(elem_size != 0, "capacity overflow");

        let align = mem::align_of::<T>();
        let layout;

        unsafe {
            let (new_cap, ptr) = if self.cap == 0 {
                // should be able to safely unwrap here, can't fail with given parameters
                layout = alloc::Layout::from_size_align(elem_size, align).unwrap();
                let ptr = alloc::alloc(layout);
                (1, ptr)
            } else {
                // we can assume self.cap < isize::MAX
                let new_cap = self.cap * 2;
                // can't overflow due to previously allocating this
                let old_num_bytes = self.cap * elem_size;

                // check that the new allocation doesn't exceed `isize::MAX`
                assert!(old_num_bytes <= (::std::isize::MAX as usize) / 2, "capacity overflow");
                // again, can't fail with the given parameters
                layout = alloc::Layout::from_size_align(old_num_bytes, align).unwrap();
                let ptr = alloc::realloc(self.ptr.as_ptr() as *mut _,
                                         layout, new_cap * elem_size);
                (new_cap, ptr)
            };

            // if alloc or realloc fail we'll get null back
            if ptr.is_null() { alloc::handle_alloc_error(layout) }

            // unwrap can't fail because we already handle null pointers
            self.ptr = NonNull::new(ptr as *mut _).unwrap();
            self.cap = new_cap;
        }
    }
}

impl<T> Drop for RawVec<T> {
    fn drop(&mut self) {
        let elem_size = mem::size_of::<T>();
        if self.cap != 0 && elem_size != 0 {
            let align = mem::align_of::<T>();
            let elem_size = mem::size_of::<T>();
            let layout = alloc::Layout::from_size_align(self.cap * elem_size, align).unwrap();
            unsafe {
                alloc::dealloc(self.ptr.as_ptr() as *mut _, layout);
            }
        }
    }
}

#[cfg(test)]
mod tests {
    use super::RawVec;

    #[test]
    fn grow() {
        let mut vec: RawVec<i32> = RawVec::new();
        assert_eq!(vec.cap, 0);
        vec.grow();
        assert_eq!(vec.cap, 1);
        vec.grow();
        assert_eq!(vec.cap, 2);
    }

    #[test]
    fn zst() {
        struct Nothing;

        let mut vec: RawVec<Nothing> = RawVec::new();
        assert_eq!(vec.cap, !0);
    }
}